﻿-- no druid stats if no calculations file loaded
if not DruidStats then
	return;
end

-- startup stuff that needs local variables
do -- some setup stuff
	
	local s = DruidStats;
	local BI = LibStub("LibBabble-Inventory-3.0"):GetLookupTable();
	-- Valid item sub-types
	s.validSubTypes = {
		[BI["Leather"]]          = true,
		[BI["Cloth"]]            = true,
		[BI["Daggers"]]          = true,
		[BI["Fist Weapons"]]     = true,
		[BI["One-Handed Maces"]] = true,
		[BI["Staves"]]           = true,
		[BI["Two-Handed Maces"]] = true,
		[BI["Miscellaneous"]]    = true,
		[BI["Other"]]            = true, -- yay for Alchemist's Stone
		
		-- Elixirs and Food effects
		[BI["Elixir"]]           = true,
		[BI["Flask"]]            = true,
		[BI["Food & Drink"]]     = true,
	};
	-- Valid item types (they will be valid no matter their subtype)
	s.validTypes = {
		[BI["Gem"]]              = true,
	};
	-- Items that proved to cause errors in specific locales
	-- TODO: implement this
	s.blacklist = {
		[28109] = "deDE",
	};
	
	-- Item ID's of special items
	s.specialItems = {
		-- Elixirs & Flasks
		[33208] = { -- Flask of Chromatic Wonder
			["STR"] = 18,
			["AGI"] = 18,
			["STA"] = 18,
			["INT"] = 18,
			["SPI"] = 18,
		},

		-- Food & Drink
		[27663] = { -- Blackened Sporefish
			["MANA_REG"] = 8,
			["STA"] = 20,
		},
		[27664] = { -- Grilled Mudfish
			["AGI"] = 20,
			["SPI"] = 20,
		},
		[33052] = { -- Fisherman's Feast
			["SPI"] = 20,
			["STA"] = 30,
		},
		[34411] = { -- Hot Apple Cider
			["SPI"] = 20,
			["STA"] = 20,
		},
		[33872] = { -- Spicy Hot Talbuk
			["SPI"] = 20,
			["MELEE_HIT_RATING"] = 20,
		},
		[33825] = { -- Skullfish Soup
			["SPI"] = 20,
			["SPELL_CRIT_RATING"] = 20,
		},
	};
	-- duplicate stats
	s.specialItems[27667] = s.specialItems[33052]; -- Spicy Crawdad
	-- Useless:
	-- Hot Buttered Trout
	-- Stewed Trout

	s.forms = {
		["No Form"]             = 1,
		[(GetSpellInfo(5487))]  = 2, -- Bear Form
		[(GetSpellInfo(9634))]  = 2, -- Dire Bear Form
		[(GetSpellInfo(768))]   = 3, -- Cat Form
		[(GetSpellInfo(33891))] = 4, -- Tree of Life
		[(GetSpellInfo(24858))] = 5, -- Moonkin
		-- unused: default to caster form
		[(GetSpellInfo(783))]   = 1, -- Travel Form
		[(GetSpellInfo(1066))]  = 1, -- Aquatic Form
		[(GetSpellInfo(33943))] = 1, -- Flight Form
		[(GetSpellInfo(40120))] = 1, -- Swift Flight Form
	};
end

DruidStats.TooltipList = {
	-- Blizzard UI
	"ItemRefTooltip", "GameTooltip",
	"ShoppingTooltip1", "ShoppingTooltip2", "ShoppingTooltip3",
	
	-- EquipCompare
	"ComparisonTooltip1", "ComparisonTooltip2",
	
	-- EQCompare
	"EQCompareTooltip1", "EQCompareTooltip2",
	
	-- LinkWrangler
	"IRR_ItemRefTooltip1", "IRR_ItemCompTooltip1", "IRR_ItemCompTool11",
	"IRR_ItemRefTooltip2", "IRR_ItemCompTooltip2", "IRR_ItemCompTool12",
	"IRR_ItemRefTooltip3", "IRR_ItemCompTooltip3", "IRR_ItemCompTool13",
	"IRR_ItemRefTooltip4", "IRR_ItemCompTooltip4", "IRR_ItemCompTool14",
	"IRR_ItemRefTooltip5", "IRR_ItemCompTooltip5", "IRR_ItemCompTool15",
	
	-- MultiTips
	"ItemRefTooltip2", "ItemRefTooltip3", "ItemRefTooltip4", "ItemRefTooltip5",
	
	-- Armory
	"ArmoryComparisonTooltip1", "ArmoryComparisonTooltip2",
	
	-- AtlasLoot
	"AtlasLootTooltip",
}

DruidStats.pd = {}; -- Player data
DruidStats.itemCache = {};

function DruidStats:updatePlayerData()
	local pd = self.pd;
	-- get the shapeshift form...
	pd.form = GetShapeshiftForm(true);
	
	if GetShapeshiftForm(true) > 0 then
		local _, name = GetShapeshiftFormInfo(GetShapeshiftForm(true));
		pd.realform = self.forms[name];
		pd.formname = name;
	else
		pd.realform = 1;
		pd.formname = "No Form";
	end
	
	local x = pd.realform or 0;
	if not self.itemCache[x] then
		self.itemCache[x] = setmetatable({}, {__mode = "kv"});
	end
	-- /run do local _, n = GetShapeshiftFormInfo(1); print(n); end
	-- /run print(GetShapeshiftForm(true).." - ".. GetShapeshiftForm());
	_,_,_,_, pd.moonkin   = GetTalentInfo(1, 18); -- Moonkin
	_,_,_,_, pd.lg        = GetTalentInfo(1, 12); -- Lunar Guidance
	_,_,_,_, pd.ds        = GetTalentInfo(1, 15); -- Dreamstate
	_,_,_,_, pd.bop       = GetTalentInfo(1, 17); -- Balance of Power
	_,_,_,_, pd.imk       = GetTalentInfo(1, 19); -- Improved Moonkin
	_,_,_,_, pd.iff       = GetTalentInfo(1, 20); -- Improved Faerie Fire
	_,_,_,_, pd.eam       = GetTalentInfo(1, 27); -- Earth and Moon
	
	_,_,_,_, pd.hotw      = GetTalentInfo(2, 17); -- Heart of the Wild
	_,_,_,_, pd.sotf      = GetTalentInfo(2, 18); -- Survival of the Fittest
	_,_,_,_, pd.th        = GetTalentInfo(2, 5 ); -- Thick Hide
	_,_,_,_, pd.ni        = GetTalentInfo(2, 15); -- Nurturing Instinct
	_,_,_,_, pd.ps        = GetTalentInfo(2, 10); -- Predatory Strikes
	_,_,_,_, pd.pp        = GetTalentInfo(2, 12); -- Primal Precision
	_,_,_,_, pd.potp      = GetTalentInfo(2, 22); -- Protector of the Pack
	
	_,_,_,_, pd.nat       = GetTalentInfo(3, 4 ); -- Naturalist
	_,_,_,_, pd.intensity = GetTalentInfo(3, 7 ); -- Intensity
	_,_,_,_, pd.masterss  = GetTalentInfo(3, 9 ); -- Master Shapeshifter
	_,_,_,_, pd.ls        = GetTalentInfo(3, 17); -- Living Spirit (15% increase to spirit)
	_,_,_,_, pd.np        = GetTalentInfo(3, 19); -- Natural Perfection
	_,_,_,_, pd.tol       = GetTalentInfo(3, 23); -- Tree of Life
	_,_,_,_, pd.itol      = GetTalentInfo(3, 24); -- Improved Tree of Life
	pd.level = UnitLevel("player");
	
	self:calculateTalentFactors(pd);
end

function DruidStats:calculateTalentFactors(pd)
	-- FERAL
	-- intellect factor for Heart of the Wild
	pd.f_HOTW_INT = 1 + (pd.hotw * 0.04);
	-- stamina factor for Heart of the Wild in BEAR FORM
	pd.f_HOTW_STA = 1 + (pd.hotw * 0.04);
	-- attack power factor for Heart of the Wild in CAT FORM
	pd.f_HOTW_AP  = 1 + (pd.hotw * 0.02);
	-- survival of the fittest factor
	pd.f_SOTF     = 1 + (pd.sotf * 0.02);
	-- survival of the fittest armor factor
	pd.f_SOTF_AR  = 1 + (pd.sotf * 0.22);
	-- nurturing instinct outgoing heal factor
	pd.f_NI       = pd.ni * 0.35;
	-- nurturing instinct incoming heal factor
	pd.f_NI_IN    = pd.ni * 0.10;
	-- predatory strikes, weapon attack power modifier
	pd.f_PS       = ({1.07, 1.14, 1.20, [0] = 1})[pd.ps];
	-- Thick hide
	if pd.th > 0 then
		pd.f_TH = 1.01 + 0.03*pd.th;
	else
		pd.f_TH = 1;
	end
	-- Protector of the Pack AP factor
	pd.f_POTP_AP  = pd.potp * 0.02;
	-- Predatory Instincts
	pd.f_PI       = ({1.03, 1.07, 1.1, [0] = 1})[pd.pi];
	

	-- lunar guidance factor
	pd.f_LG       = pd.lg * 0.04;
	-- dreamstate
	pd.f_DS       = ({1.04, 1.07, 1.1, [0] = 1})[pd.ds];
	-- Balance of Power
	pd.f_BOP      = pd.bop * 2;
	-- Improved Moonkin Form Spelldamage component
	pd.f_IMK_SD   = 1 + (pd.imk * 0.05);
	-- Improved Moonkin Form Haste component
	pd.f_IMK_HA   = 1 + (pd.imk * 0.01);
	-- Improved Faerie Fire (both components)
	pd.f_IFF      = 1 + (pd.iff * 0.01);
	-- Earth and Moon (Spelldamage Component)
	pd.f_EAM_SD   = 1 + (pd.eam * 0.01);
	
	-- Naturalist
	pd.f_NAT      = 1 + (pd.nat * 0.02);
	-- Intensity
	pd.f_INT      = (pd.intensity * 0.1);
	-- Master Shapeshifter (Bear/MK/Tree modifier)
	pd.f_MASTERSSA= 1 + (pd.masterss * 0.02);
	-- Master Shapeshifter (Cat modifier)
	pd.f_MASTERSSB= pd.masterss * 2;
	-- Natural perfection
	pd.f_NP       = pd.np;
	-- living spirit
	pd.f_LS       = 1 + (pd.ls * 0.05);
	-- Improved ToL (Armor Component)
	pd.f_ITOL_AR  = ({1.33, 1.66, 2, [0] = 1})[pd.itol];
	-- Improved ToL (Healing Power Component)
	pd.f_ITOL_HP  = pd.itol * 0.05;
end

function DruidStats.SetItem(tooltip)
	--print(tooltip:GetName(), select(2, tooltip:GetItem()));
	local self = DruidStats;
	if not DruidStats_Settings.enabled then
		return;
	end
	-- Get the item
	local _, link = tooltip:GetItem();
	if not link then
		return;
	end
	-- Check if the item is a usable type
	local _,_,_,_,_, Type, subType = GetItemInfo(link);
	if not self.validSubTypes[subType] and not self.validTypes[Type] then
		return;
	end
	-- Check the cache
--	debugprofilestart();
	local iText = "";
	if self.itemCache[self.pd.realform][link] then
		iText = self.itemCache[self.pd.realform][link];
	else
		-- Special items
		local bonuses;
		local itemid  = tonumber(strmatch(link, "Hitem:(%d+):"));
		if self.specialItems[itemid] ~= nil then
			bonuses = self.specialItems[itemid];
		else
			bonuses = StatLogic:GetSum(link) or {};
		end
		-- parse the data
		for stat, s in pairs(self.sstats[self.pd.realform]) do
			local val, v2, v3 = s.f(bonuses); -- Call the function for the stat
			if val and val ~= 0 then
				iText = iText..string.format(s.p, val, v2, v3);
			end
		end
		-- save to cache
		self.itemCache[self.pd.realform][link] = iText;
	end

	if iText ~= "" then
		tooltip:AddLine("|n<< Druid Stats - "..self.pd.formname.." >>");
		tooltip:AddLine(iText);
	end
--	print("DS-new: "..debugprofilestop().."ms");
end

--[[ Slash Commands ]]--
DruidStats.SlashHandler = function(msg)
	if msg == "on" then
		print("DruidStats: Enabled")
		DruidStats_Settings.enabled = true;
	elseif msg == "off" then
		print("DruidStats: Disabled")
		DruidStats_Settings.enabled = false;
	elseif msg == "unfake" then
		DruidStats:updatePlayerData();
		DruidStats.itemCache = {};
	elseif string.match(msg, "^fake%s") then
		local pd = DruidStats.pd;
		for w in string.gmatch(msg, "[^%s]+") do
			if w == "new" then
				DruidStats:updatePlayerData();
			elseif w == "bear" then
				pd.realform = 2;
				pd.formname = "Fake BEAR";
			elseif w == "cat" then
				pd.realform = 3;
				pd.formname = "Fake CAT";
			elseif w == "tol" then
				pd.realform = 4;
				pd.formname = "Fake TOL";
			elseif w == "mk" or w == "moonkin" then
				pd.realform = 5;
				pd.formname = "Fake MK";
			elseif w == "hotw" then
				pd.hotw = 5;
			elseif w == "sotf" then
				pd.sotf = 3;
			elseif w == "int" then
				pd.intensity = 3;
			elseif w == "ni" then
				pd.ni = 2;
			elseif w == "ds" then
				pd.ds = 3;
			elseif w == "lg" then
				pd.lg = 3;
			elseif w == "ls" then
				pd.ls = 3;
			elseif w == "th" then
				pd.th = 3;
			end
		end
		DruidStats:calculateTalentFactors(DruidStats.pd);
		DruidStats.itemCache = {};
		setmetatable(DruidStats.itemCache, {__mode = "kv"}); -- weak table to enable garbage collection
	elseif msg == "bench" then
		-- DruidStats from global space vs. DruidStats from local space
		print("Bench: Access table via global space VS reference it and access it via local space");
		do
			local m = 1;
			local limit = 10000;
			
			while m < 10 do
				local k = DruidStats;
				local l,g = 0;
				k.benchI = 0;
				debugprofilestart();
				while k.benchI < limit do
					local i = DruidStats;
					local j = 0;
					while i.benchI < limit and j < m do
						i.benchI = i.benchI+1;
						j = j + 1;
					end
				end
				l = debugprofilestop();
				
				k.benchI = 0;
				debugprofilestart();
				while DruidStats.benchI < limit do
					local j = 0;
					while DruidStats.benchI < limit and j < m do
						DruidStats.benchI = DruidStats.benchI+1;
						j = j + 1;
					end
				end
				g = debugprofilestop();

				print("L/G: "..l/g);
				print("Local: "..l.."ms - "..(3*m).." calls per localized loop");
				print("Global: "..g.."ms - "..(3*m).." calls per loop");
				
				m = m+1;
			end
		end
	else
		print("DruidStats: Invalid command, usage: /druidstats on, /druidstats off")
	end
end

SlashCmdList["DRUID_STATS"] = DruidStats.SlashHandler;
SLASH_DRUID_STATS1 = "/druidstats";
SLASH_DRUID_STATS2 = "/ds";

DruidStats_Settings = {};
DruidStats_Settings.debug = false;
DruidStats_Settings.enabled = true;

local updateForm = function()
	local pd = DruidStats.pd;
	-- get the shapeshift form...
	pd.form = GetShapeshiftForm(true);
	
	if GetShapeshiftForm(true) > 0 then
		local _, name = GetShapeshiftFormInfo(GetShapeshiftForm(true));
		pd.realform = DruidStats.forms[name];
		pd.formname = name;
	else
		pd.realform = 1;
		pd.formname = "No Form";
	end
	
	if not DruidStats.itemCache[pd.realform] then
		local x = pd.realform;
		DruidStats.itemCache[pd.realform or 0] = {};
		setmetatable(DruidStats.itemCache[pd.realform or 0], {__mode = "kv"});
	end
end

local clearCache = function()
	-- Update the player data
	DruidStats:updatePlayerData();
	-- Reset the cache
	DruidStats.itemCache = {};
	DruidStats.itemCache[DruidStats.pd.realform or 0] = {};
	setmetatable(DruidStats.itemCache[DruidStats.pd.realform or 0], {__mode = "kv"});
end

local startup = function(self, event, name)
	if name == "DruidStats" then
		clearCache(self, event, name);
		--print("|cffa335ee|Hitem:21357:0:0:0:0:0:0:0:0|h[Genesis Vest]|h|r");
	end
	--[[ Hook DruidStats into various tool tips ]]--
	for k, name in ipairs(DruidStats.TooltipList) do
		local tooltip = _G[name];
		if tooltip and tooltip:HasScript("OnTooltipSetItem") then
			-- Hook this script
			if tooltip:GetScript("OnTooltipSetItem") then
				tooltip:HookScript("OnTooltipSetItem", DruidStats.SetItem);
			else
				tooltip:SetScript("OnTooltipSetItem", DruidStats.SetItem);
			end
			-- remove the frame from the list so I won't hook it more than once
			DruidStats.TooltipList[k] = nil;
			--print("[DS] Hooked", tooltip:GetName());
		end
	end
end

-- dynamically create an event catcher frame
local DruidStatsFrame1 = CreateFrame("Frame", nil, UIParent);
DruidStatsFrame1:SetScript("OnEvent", startup);
DruidStatsFrame1:RegisterEvent("ADDON_LOADED");

-- frame to catch events that require new parsing of a tooltip
local DruidStatsFrame2 = CreateFrame("Frame", nil, UIParent);
DruidStatsFrame2:SetScript("OnEvent", clearCache);
DruidStatsFrame2:RegisterEvent("VARIABLES_LOADED");         -- Startup
DruidStatsFrame2:RegisterEvent("CHARACTER_POINTS_CHANGED"); -- Talents changed
DruidStatsFrame2:RegisterEvent("PLAYER_LEVEL_UP");          -- Leveling
DruidStatsFrame2:RegisterEvent("UPDATE_SHAPESHIFT_FORMS");  -- When changing the AVAILABLE set of stances/forms
DruidStatsFrame2:RegisterEvent("UNIT_INVENTORY_CHANGED");   -- When (un)equipping items

-- frame to catch events that only change the form of the player
local DruidStatsFrame3 = CreateFrame("Frame", nil, UIParent);
DruidStatsFrame3:SetScript("OnEvent", updateForm);
DruidStatsFrame3:RegisterEvent("UPDATE_SHAPESHIFT_FORM");   -- When changing stances/forms
